﻿{ delQuery v0.1 - jQuery like library for Delphi VCL.
  http://bitbucket.org/utku/delquery/

  (c) 2009 M.Utku Karataş. <m.utku.k@gmail.com>
  This library is freely distributable under the terms of the MIT license.
  Inspired by jQuery by John Resig and the jQuery Team - jquery.com.

  TODO:
  - mouseenter-mouseleave has bugs
  - .prop(array of props)
  - .toggle(array of handlers)
  - insert all doc comments
  - better mem-leak proof TdelQuery instances.
  - better RTTI intergration with .prop
}
unit delquery;

interface

uses
  Classes, SysUtils, Controls, Windows, Messages, Graphics, Forms, Generics.Collections,
  ExtCtrls, Rtti;

type
  TdelQuery = class;

  { All delQuery event handlers are passed this construct as parameter. }
  dqevent = packed record
  private
    function GetThis: TdelQuery;
  public
    sender: TControl;                      { Reference to the original Sender parameter of events. }
    prevent: boolean;                      { Set this to true to prevent the default event handlers trigger }
    property this: TdelQuery read GetThis; { So that you can use "event.this" instead of
                                             "dq(event.sender)" every time. }
    case Integer of
      0: ( { mouse events }
        button: TMouseButton;
        shift: TShiftState;
        x, y: Integer;);
      1: ( { key events }
        {shift}
        key: Word;);
      2: ( {resize events }
        NewWidth: integer;
        NewHeight: integer;
        Resize: boolean;);
  end;

  TdqEachCallback = reference to procedure(obj: TControl);
  TdqEventHandler = reference to procedure(event: dqevent);
  TdqAnimationDoneCallback = reference to procedure(event: dqevent);
  TdqEasingFunc = function(x, t, b, c, d: double): double;

  { The all-mighty class providing every thing. Basically it is a list that holds control references }
  TdelQuery = class(TComponent)
  private
    fList: TList<TControl>;
    procedure AddControl(aControl: TControl);
    { event stubs }
    procedure OnClickStub(Sender: TObject);
    procedure OnEnterStub(Sender: TObject);
    procedure OnExitStub(Sender: TObject);
    procedure OnMouseDownStub(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure OnMouseUpStub(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure OnMouseMoveStub(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure OnKeyDownStub(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnKeyUpStub(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnKeyPressStub(Sender: TObject; var Key: Char);
    procedure OnResizeStub(Sender: TObject);
    procedure OnCanResizeStub(Sender: TObject; var NewWidth, NewHeight: Integer; var Resize: Boolean);
    function GetCount: integer;
    function GetItems(Index: integer): TControl;
  protected
    fFxTimer: TTimer;
    function GetFx: TTimer;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property FxTimer: TTimer read GetFx;
    function DoAnimate(obj: TControl; props: array of string; duration: string;
      easing: TdqEasingFunc = nil; callback: TdqAnimationDoneCallback = nil): TdelQuery;
  public
    constructor Create; reintroduce;
    destructor Destroy; override;

    { core }
    function each(callback: TdqEachCallback): TdelQuery;
    property count: integer read GetCount; // jQuery equivalent: size
    property get[Index: integer]: TControl read GetItems; default;

    { properties }
    function prop(const key: string): TValue; overload; // jQuery equivalent: jQuery.attr
    function prop(const key: string; value: TValue): TdelQuery; overload;

    { events }
    function enter(fn: TdqEventHandler): TdelQuery;
    function exit_(fn: TdqEventHandler): TdelQuery;
    function click(fn: TdqEventHandler): TdelQuery;
    function mouseenter(fn: TdqEventHandler): TdelQuery;
    function mouseleave(fn: TdqEventHandler): TdelQuery;
    function mouseup(fn: TdqEventHandler): TdelQuery;
    function mousedown(fn: TdqEventHandler): TdelQuery;
    function mousemove(fn: TdqEventHandler): TdelQuery;
    function keydown(fn: TdqEventHandler): TdelQuery;
    function keyup(fn: TdqEventHandler): TdelQuery;
    function keypress(fn: TdqEventHandler): TdelQuery;
    function oncanresize(fn: TdqEventHandler): TdelQuery;
    function onresize(fn: TdqEventHandler): TdelQuery;

    function toggle(fn, fn2: TdqEventHandler): TdelQuery;

    { effects }
    function Animate(props: array of const;
                     duration: string;
                     easing: TdqEasingFunc = nil;
                     callback: TdqAnimationDoneCallback = nil): TdelQuery;
    //todo:  procedure SlideUp(speed, callback);
    // procedure SlideDown(speed, callback);
    // procedure SlideToggle(speed, callback);
  end;

function dq(aClass: TClass; Parent: TWinControl; Greedy: boolean = False): TdelQuery; overload;
function dq(aControl: TControl): TdelQuery; overload;
function dq(Controls: array of TControl): TdelQuery; overload;

function ease_lineer(x, t, b, c, d: double): double;
function ease_swing(x, t, b, c, d: double): double;
function ease_outbounce(x, t, b, c, d: double): double;
function ease_inelastic(x, t, b, c, d: double): double;

implementation

uses
  Types, Math, Dialogs, DateUtils;

{ -------------------------------------------------------------------------------------------------------------------- }
{ Internal structs of use }

type
  TWinControlAccessHack = class(TWinControl);

  TAnimationData = class;

  { Too bad class helpers can't attach fields to a class, this is gonna be a weird way
    but every control needs to be attached a copy of ExtraData }
  ExtraData = class
  private
    fSelf: TControl;
    AnimationQueue: TObjectQueue<TAnimationData>;
    fDQ: TdelQuery;
    fThis: TdelQuery;
    fOldWindowProc: TWndMethod;
    AlreadyInElement: boolean;
    procedure OnWindowProc(var Message: TMessage);
  private
    { New event to trrigger }        { The old event before assignment }
    EnterFn: TdqEventHandler;        OldOnEnter: TNotifyEvent;
    ExitFn: TdqEventHandler;         OldOnExit: TNotifyEvent;
    ClickFn: TdqEventHandler;        OldOnClick: TNotifyEvent;
    MouseEnterFn: TdqEventHandler;   //OldOnMouseEnter: TNotifyEvent;
    MouseLeaveFn: TdqEventHandler;   //OldOnMouseLeave: TNotifyEvent;
    MouseUpFn: TdqEventHandler;      OldOnMouseUp: TMouseEvent;
    MouseDownFn: TdqEventHandler;    OldOnMouseDown: TMouseEvent;
    MouseMoveFn: TdqEventHandler;    OldOnMouseMove: TMouseMoveEvent;
    KeyDownFn: TdqEventHandler;      OldOnKeyDown: TKeyEvent;
    KeyUpFn: TdqEventHandler;        OldOnKeyUp: TKeyEvent;
    KeyPressFn: TdqEventHandler;     OldOnKeyPress: TKeyPressEvent;
    CanResizeFn: TdqEventHandler;    OldOnCanResize: Controls.TCanResizeEvent;
    ResizeFn: TdqEventHandler;       OldOnResize: TNotifyEvent;
  public
    class function for_(obj: TObject): ExtraData;
  end;

  TAnimationData = class
  public
    props: array of string;
    duration: string;
    easing: TdqEasingFunc;
    callback: TdqAnimationDoneCallback;
  end;

  { easy to use timer }
  TFxOnStop = reference to procedure;
  TFxOnStep = reference to procedure(var stop: boolean);

  TFxTimer = class(TTimer)
  protected
    fOnStep: TFxOnStep;
    fOnStop: TFxOnStop;
    procedure OnTimerStub(Sender: TObject);
  public
    procedure FireUp(OnStep: TFxOnStep; OnStop: TFxOnStop);
  end;

var
  ExtraDataRegistry: TDictionary<TObject, ExtraData>;

class function ExtraData.for_(obj: TObject): ExtraData;
begin
  if not ExtraDataRegistry.TryGetValue(obj, Result) then
  begin
    Result := ExtraData.Create;
    Result.fSelf := obj as TControl;
    Result.AnimationQueue := TObjectQueue<TAnimationData>.Create;
    Result.AnimationQueue.OwnsObjects := False;
    ExtraDataRegistry.Add(obj, Result);
  end;
end;

procedure TFxTimer.FireUp(OnStep: TFxOnStep; OnStop: TFxOnStop);
begin
  Self.fOnStep := OnStep;
  Self.fOnStop := OnStop;
  Self.OnTimer := OnTimerStub;
  Self.Interval := 13;
  Self.Enabled := True;
end;

procedure TFxTimer.OnTimerStub;
var
  stop: boolean;
begin
  stop := false;
  Self.fOnStep(stop);
  if stop then
  begin
    Self.Enabled := false;
    Self.fOnStop();
  end;
end;

function dqEvent.GetThis: TdelQuery;
var data: ExtraData;
begin
  data := ExtraData.for_(sender);
  if data.fThis = nil then
  begin
    Result := TdelQuery.Create;
    Result.AddControl(sender);
  end
  else
    Result := data.fThis;
end;

procedure log(const S: string);
begin
  OutputDebugString(PWideChar(S));
end;

function ObjGetInitialValue(Obj: TObject; const Prop: string): integer;
begin
  Result := TRttiContext.Create.GetType(Obj.ClassType).GetProperty(Prop).GetValue(obj).AsInteger;
end;

procedure ObjSetValue(Obj: TObject; const Prop: string; Value: TValue); overload;
var
  dotPos: integer;
  curProp, curProp2: string;
  curRProp: TRttiProperty;
begin
  dotPos := Pos('.', Prop);
  if dotPos = 0 then
    TRttiContext.Create.GetType(Obj.ClassType).GetProperty(Prop).SetValue(Obj, Value)
  else
  begin
    { support calls like 'font.color' }
    // todo: support more depth
    curProp := Copy(Prop, 1, dotPos - 1);
    curRProp := TRttiContext.Create.GetType(Obj.ClassType).GetProperty(curProp);
    curProp2 := Copy(Prop, dotPos + 1, MaxInt);
    if curRProp.PropertyType is TRttiInstanceType then
      TRttiContext.Create.GetType(curRProp.GetValue(Obj).AsObject.ClassType).
        GetProperty(curProp2).SetValue(curRProp.GetValue(Obj).AsObject, Value);
  end;
end;

function ObjGetTargetValue(const S: string; InitialValue: integer): integer;
begin
  try
    if Copy(S, 1, 2) = '+=' then
      Result := StrToInt(Copy(S, 3, Maxint)) + InitialValue
    else if Copy(S, 1, 2) = '-=' then
      Result := InitialValue - StrToInt(Copy(S, 3, Maxint))
    else
      Result := StrToInt(S);
  except
    raise Exception.CreateFmt('Unrecognized param value "%s"', [S]);
  end;
end;

{ Delphi doesn't let you use open arrays out of function context but we need
  to store given values. This function converts an open array to a string array
  for further use.}
type TStringArray = array of string;
function StrArrayFromOpenArray(openArr: array of const): TStringArray;
var
  i: integer;
  v: PVarRec;
begin
  SetLength(Result, Length(openArr));
  for i := low(openArr) to high(openArr) do
  begin
    v := @openArr[i];
    case v.VType of
      vtString: Result[i] := string(v.VString);
      vtUnicodeString: Result[i] := string(v.VString);
      vtInteger: Result[i] := IntToStr(v.VInteger);
    else
      raise Exception.Create('xxx unsupported property type to use in animations!');
    end;
  end;
end;

{ -------------------------------------------------------------------------------------------------------------------- }
{ Functions for easing effects

  See http://jquery-ui.googlecode.com/svn/trunk/demos/effect/easing.html for the effects in use.
  Can add more functions from http://gsgd.co.uk/sandbox/jquery/easing/jquery.easing.1.3.js }

function ease_lineer(x, t, b, c, d: double): double;
begin
  Result := b + c * x;
end;

function ease_swing(x, t, b, c, d: double): double;
begin
  Result := ((-cos(x * PI) / 2) + 0.5) * c + b;
end;

function ease_outbounce(x, t, b, c, d: double): double;
begin
  t := t / d;
  if (t < (1 / 2.75)) then
    Result := c * (7.5625 * t * t) + b
  else
  if (t < (2 / 2.75)) then
  begin
    t := t - (1.5 / 2.75);
    Result := c * (7.5625 * t * t + 0.75) + b;
  end
  else if (t < (2.5 / 2.75)) then
  begin
    t := t - (2.25 / 2.75);
    Result := c * (7.5625 * t * t + 0.9375) + b;
  end
  else
  begin
    t := t - (2.625 / 2.75);
    Result := c * (7.5625 * t * t + 0.984375) + b;
  end;
end;

function ease_inelastic(x, t, b, c, d: double): double;
var
  a,s,p: double;
begin
  p := 0;
  a := c;
  if (t = 0) then EXIT(b);
  t := t / d;
  if (t = 1) then EXIT(b+c);
  if p = 0 then
    p := d * 0.3;
  if (a < abs(c)) then
  begin
    a := c;
    s := p/4;
  end
  else
    s := p / (2 * PI) * ArcSin(c / a);
  t := t - 1;
  Result := -(a * Power(2, 10 * t) * sin((t * d - s) * (2 * PI) / p)) + b;
end;

{ -------------------------------------------------------------------------------------------------------------------- }
{ TdelQuery }

constructor TdelQuery.Create;
begin
  fList := TList<TControl>.Create;
end;

destructor TdelQuery.Destroy;
begin
  FreeAndNil(fFxTimer);
  FreeAndNil(fList);
end;

procedure TdelQuery.AddControl(aControl: TControl);
begin
  fList.Add(aControl);
  aControl.FreeNotification(self);
end;

function TdelQuery.GetCount: integer;
begin
  Result := fList.Count
end;

function TdelQuery.GetFx: TTimer;
begin
  if fFxTimer = nil then
    fFxTimer := TFxTimer.Create(nil);
  Result := TFxTimer(fFxTimer);
end;

function TdelQuery.GetItems(Index: integer): TControl;
begin
  Result := fList[Index];
end;

procedure TdelQuery.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    fList.Delete(fList.IndexOf(TControl(AComponent)));
    if fList.Count = 0 then
      Destroy; // log('all controls of delQuery are gone and so is the delQuery object..');
  end;
end;

function TdelQuery.each(callback: TdqEachCallback): TdelQuery;

  { }

var
  i: integer;
begin
  Result := Self;
  for i := 0 to fList.Count - 1 do
    callback(fList[i]);
end;

function TdelQuery.prop(const key: string): TValue;

  { Access a property on the first element (first control in delQuery instance). }

begin
  if count = 0 then
    result := TValue.Empty
  else
    Result := TRttiContext.Create.GetType(get[0].ClassType).GetProperty(key).GetValue(get[0]);
end;

function TdelQuery.prop(const key: string; value: TValue): TdelQuery;

  { Set the property given by "key" to the "value" on all controls.
    Using TValue gives us implicit conversion for most of the types }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ObjSetValue(obj, key, value);
            end);
end;

{??? function TdelQuery.prop(const props: array of const): TdelQuery;
var i: integer;
begin
  if Odd(Length(props)) then raise Exception.Create('Every property must match a value!');
  Result := Self;
  for i := 1 to (Length(props) div 2) do
  begin
    prop(string(props[i*2-1].VUnicodeString), '');
  end;
end; ???}

type TStepFunc = reference to procedure;

function TdelQuery.DoAnimate(obj: TControl; props: array of string; duration: string;
  easing: TdqEasingFunc; callback: TdqAnimationDoneCallback): TdelQuery;
var
  state, pos: double;
  durationMS, i: integer;
  startTime: TDateTime;
  callback_: TdqAnimationDoneCallback; // clone for capturing
  props_: array of string; // clone for capturing
  step, lastStep: TStepFunc;
  currentValues, startValues, endValues: TArray<integer>;

  procedure FixUpDuration;
  begin
    durationMS := StrToIntDef(duration, -1);
    if durationMS = -1 then
      if duration = 'slow' then
        durationMS := 600
      else if duration = 'default' then
        durationMS := 400
      else if duration = 'fast' then
        durationMS := 200;
  end;

  function PushToQueue: boolean;
  var
    data: TAnimationData;
    i: integer;
  begin
    Result := False;
    if obj.Tag = 1 then // there's already an animation
    begin
      data := TAnimationData.Create;
      SetLength(data.props, Length(props)); // silly pascal now letting array of strings autocopy
      for i := 0 to Length(props) - 1 do
        data.props[i] := props[i];
      data.duration := duration;
      data.easing := easing;
      data.callback := callback;
      Extradata.for_(obj).AnimationQueue.Enqueue(data);
      //log(Format('PUSH animation with %s, %s', [props[0], props[1]]));
      Result := True;
    end;
  end;

  procedure InitFxVars;
  var i: integer;
  begin
    SetLength(currentValues, Length(props_) div 2);
    SetLength(startValues, Length(props_) div 2);
    SetLength(endValues, Length(props_) div 2);
    for i := 0 to (Length(props_) div 2) - 1 do
    begin
      startValues[i] := ObjGetInitialValue(obj, props_[i*2]);
      endValues[i] := ObjGetTargetValue(props[i*2+1], startValues[i]);
    end;
  end;

begin
  //log(Format('REQUEST animation with %s, %s', [props[0], props[1]]));
  Result := Self;
  if PushToQueue then EXIT;
  obj.Tag := 1;
  //log(Format('IN animation with %s, %s', [props[0], props[1]]));

  SetLength(props_, Length(props));
  for i := 0 to High(props) do props_[i] := props[i];
  callback_ := callback;

  FixUpDuration;
  if @easing = nil then
    easing := ease_lineer;
  InitFxVars;
  startTime := Now;

  step := procedure
          var i: integer;
          begin
            for i := 0 to (Length(props_) div 2) - 1 do
            begin
             currentValues[i] := trunc(startValues[i] + ((endValues[i] - startValues[i]) * pos));
              ObjSetValue(obj, props_[i*2], currentValues[i]);
            end;
          end;

  lastStep := procedure
              var i: integer;
              begin
                for i := 0 to (Length(props_) div 2) - 1 do
                begin
                  ObjSetValue(obj, props_[i*2], endValues[i]);
                end;
              end;

  TFxTimer(FxTimer).FireUp(
    procedure(var stop: boolean) // OnStep
    var n: integer;
    begin
      if durationMS <= MilliSecondsBetween(startTime, Now) then
      begin
        stop := True;
      end
      else
      begin
        n := MilliSecondsBetween(startTime, Now);
        state := n / durationMS;

        // Perform the easing function
        pos := easing(state, n, 0, 1, durationMS);

        // next step in the animation
        step;
      end;
    end
    ,
    procedure // OnStop
    var
      event: dqevent;
      q: TAnimationData;
    begin
      try
        obj.Tag := 0;

        // last step in the animation. (easing funcs don't complete till last bit, so force it)
        lastStep;

        // trigger done-animation callback
        if assigned(callback_) then
        begin
          event.sender := obj;
          callback_(event);
        end;
      finally
        // call the next effect in queue
        if Extradata.for_(obj).AnimationQueue.Count > 0 then
        begin
          q := Extradata.for_(obj).AnimationQueue.Peek;
          Extradata.for_(obj).AnimationQueue.Dequeue;
          DoAnimate(obj, q.props, q.duration, q.easing, q.callback);
        end;
      end;
    end
    );
end;

{ event redirections }

function TdelQuery.click(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnClick event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             begin
               ExtraData.for_(obj).OldOnClick := TWinControlAccessHack(obj).OnClick;
               TWinControlAccessHack(obj).OnClick := OnClickStub;
               ExtraData.for_(obj).ClickFn := fn;
             end );
end;

procedure TdelQuery.OnClickStub(Sender: TObject);
var event: dqevent;
begin
  event.prevent := False;
  event.sender := Sender as TControl;
  ExtraData.for_(Sender).ClickFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnClick <> nil then
      ExtraData.for_(Sender).OldOnClick(Sender);
end;

function TdelQuery.enter(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnEnter event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             begin
               ExtraData.for_(obj).OldOnEnter := TWinControlAccessHack(obj).OnEnter;
               TWinControlAccessHack(obj).OnEnter := OnEnterStub;
               ExtraData.for_(obj).EnterFn := fn;
             end );
end;

procedure TdelQuery.OnEnterStub(Sender: TObject);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  ExtraData.for_(Sender).EnterFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnEnter <> nil then
      ExtraData.for_(Sender).OldOnEnter(Sender);
end;

function TdelQuery.exit_(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnExit event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             begin
               ExtraData.for_(obj).OldOnExit := TWinControlAccessHack(obj).OnExit;
               TWinControlAccessHack(obj).OnExit := OnExitStub;
               ExtraData.for_(obj).ExitFn := fn;
             end );
end;

procedure TdelQuery.OnExitStub(Sender: TObject);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  ExtraData.for_(Sender).ExitFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnExit <> nil then
      ExtraData.for_(Sender).OldOnExit(Sender);
end;

function TdelQuery.mousedown(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnMouseDown event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             begin
               ExtraData.for_(obj).OldOnMouseDown := TWinControlAccessHack(obj).OnMouseDown;
               TWinControlAccessHack(obj).OnMouseDown := OnMouseDownStub;
               ExtraData.for_(obj).MouseDownFn := fn;
             end );
end;

procedure TdelQuery.OnMouseDownStub(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.button := Button;
  event.shift := Shift;
  event.x := X;
  event.y := Y;
  ExtraData.for_(Sender).MouseDownFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnMouseDown <> nil then
      ExtraData.for_(Sender).OldOnMouseDown(Sender, Button, Shift, X, Y);
end;

function TdelQuery.mouseup(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnMouseUp event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             begin
               ExtraData.for_(obj).OldOnMouseUp := TWinControlAccessHack(obj).OnMouseUp;
               TWinControlAccessHack(obj).OnMouseUp := OnMouseUpStub;
               ExtraData.for_(obj).MouseUpFn := fn;
             end );
end;

procedure TdelQuery.OnMouseUpStub(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.button := Button;
  event.shift := Shift;
  event.x := X;
  event.y := Y;
  ExtraData.for_(Sender).MouseUpFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnMouseUp <> nil then
      ExtraData.for_(Sender).OldOnMouseUp(Sender, Button, Shift, X, Y);
end;

function TdelQuery.mousemove(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnMouseMove event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ExtraData.for_(obj).OldOnMouseMove := TWinControlAccessHack(obj).OnMouseMove;
              TWinControlAccessHack(obj).OnMouseMove := OnMouseMoveStub;
              ExtraData.for_(obj).MouseMoveFn := fn;
            end);
end;

procedure TdelQuery.OnMouseMoveStub(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.shift := Shift;
  event.x := X;
  event.y := Y;
  ExtraData.for_(Sender).MouseMoveFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnMouseDown <> nil then
      ExtraData.for_(Sender).OldOnMouseMove(Sender, Shift, X, Y);
end;

procedure ExtraData.OnWindowProc(var Message: TMessage);

  function WithinElement: boolean;
  var
    c: TControl;
  begin
    c := TControl(Message.LParam);
    while c <> nil do
    begin
      if c.Parent = fSelf then EXIT(True);
      c := c.Parent;
    end;
    Result := False;
  end;

  function WithinBoundaries: boolean;
  var p: TPoint;
  begin
    p := fSelf.ScreenToClient(Mouse.CursorPos);
    Result := not (((p.X <= 0) or (p.X >= fSelf.Width)) or
                   ((p.Y <= 0) or (p.Y >= fSelf.Height)));
  end;

  procedure DoEnter;
  var event: dqevent;
  begin
    AlreadyInElement := True;
    event.Sender := fSelf as TControl;
    if assigned(MouseEnterFn) then
      MouseEnterFn(event);
  end;

  procedure DoLeave;
  var event: dqevent;
  begin
    event.Sender := fSelf as Tcontrol;
    if assigned(MouseEnterFn) then
      MouseLeaveFn(event);
    AlreadyInElement := False;
  end;

begin
  // todo: refactor this shit completely - there're some subtle bugs anyway
  // Message.LParam is 0 when the message is about fSelf, the message is about children if not.
  if Message.Msg = CM_MOUSEENTER then
  begin
    if ((Message.LParam = 0) or (WithinElement)) and not AlreadyInElement then
      DoEnter;
  end
  else
  if Message.Msg = CM_MOUSELEAVE then
  begin
    if (Message.LParam = 0) and not WithinBoundaries {xxx and not EatNextLeaveWithLparam0} then
      DoLeave;
  end;

  if assigned(fOldWindowProc) then
    fOldWindowProc(Message);
end;

function TdelQuery.mouseenter(fn: TdqEventHandler): TdelQuery;

  { xxx VCL OnMouseEnter/Leave is not enough }

begin
  Result := Self;
  Self.each( procedure(obj: TControl)
             var data: ExtraData;
             begin
               data := ExtraData.for_(obj);
//               data.OldOnMouseEnter := TWinControlAccessHack(obj).OnMouseEnter;
//               TWinControlAccessHack(obj).OnMouseEnter := OnMouseEnterStub;
               data.MouseEnterFn := fn;

               { hook the wndproc for proper mouseenter/mouseleave dispatching }
//               if @data.fOldWindowProc <> @obj.WindowProc then
//               begin
//                 data.fOldWindowProc:= obj.WindowProc;
//                 obj.WindowProc := data.OnWindowProc;
//               end;
             end );
end;

//procedure TdelQuery.OnMouseEnterStub(Sender: TObject);
//var event: dqevent;
//begin
//  event.Sender := Sender as TControl;
//  ExtraData.for_(Sender).MouseEnterFn(event);
//  if not event.prevent then
//    if @ExtraData.for_(Sender).OldOnMouseEnter <> nil then
//      ExtraData.for_(Sender).OldOnMouseEnter(Sender);
//end;

function TdelQuery.mouseleave(fn: TdqEventHandler): TdelQuery;

  { xxx VCL OnMouseEnter/Leave is not enough }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            var data: ExtraData;
            begin
              data := ExtraData.for_(obj);
              //ExtraData.for_(obj).OldOnMouseLeave := TWinControlAccessHack(obj).OnMouseLeave;
              //TWinControlAccessHack(obj).OnMouseLeave := OnMouseLeaveStub;
              data.MouseLeaveFn := fn;

               { hook the wndproc for proper mouseenter/mouseleave dispatching }
               if @data.fOldWindowProc <> @obj.WindowProc then
               begin
                 data.fOldWindowProc:= obj.WindowProc;
                 obj.WindowProc := data.OnWindowProc;
               end;
            end);
end;

//procedure TdelQuery.OnMouseLeaveStub(Sender: TObject);
//var event: dqevent;
//begin
//  event.sender := Sender as TControl;
//  if not event.prevent then
//    if @ExtraData.for_(Sender).OldOnMouseLeave <> nil then
//      ExtraData.for_(Sender).OldOnMouseLeave(Sender);
//end;

function TdelQuery.keydown(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnKeyDown event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ExtraData.for_(obj).OldOnKeyDown := TWinControlAccessHack(obj).OnKeyDown;
              TWinControlAccessHack(obj).OnKeyDown := OnKeydownStub;
              ExtraData.for_(obj).KeyDownFn := fn;
            end);
end;

procedure TdelQuery.OnKeyDownStub(Sender: TObject; var Key: Word; Shift: TShiftState);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.shift := Shift;
  event.key := Key;
  ExtraData.for_(Sender).KeyDownFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnKeyDown <> nil then
      ExtraData.for_(Sender).OldOnKeyDown(Sender, Key, Shift);
end;

function TdelQuery.keyup(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnKeyUp event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ExtraData.for_(obj).OldOnKeyUp := TWinControlAccessHack(obj).OnKeyUp;
              TWinControlAccessHack(obj).OnKeyUp := OnKeyUpStub;
              ExtraData.for_(obj).KeyUpFn := fn;
            end);
end;

procedure TdelQuery.OnKeyUpStub(Sender: TObject; var Key: Word; Shift: TShiftState);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.shift := Shift;
  event.key := Key;
  ExtraData.for_(Sender).KeyUpFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnKeyUp <> nil then
      ExtraData.for_(Sender).OldOnKeyUp(Sender, Key, Shift);
end;

function TdelQuery.keypress(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control gets OnKeyPress event. Check dqevent struct for
    extra event parameters. }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ExtraData.for_(obj).OldOnKeyPress := TWinControlAccessHack(obj).OnKeyPress;
              TWinControlAccessHack(obj).OnKeyPress := OnKeyPressStub;
              ExtraData.for_(obj).KeyPressFn := fn;
            end);
end;

procedure TdelQuery.OnKeyPressStub(Sender: TObject; var Key: Char);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.key := Word(Key);
  ExtraData.for_(Sender).KeyPressFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnKeyPress <> nil then
      ExtraData.for_(Sender).OldOnKeyPress(Sender, Key);
end;

function TdelQuery.oncanresize(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control is about to be resized. Check dqevent struct
    for extra event parameters. }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              //ExtraData.for_(obj).OldOnCanResize := TWinControlAccessHack(obj).OnCanResize;
              TWinControlAccessHack(obj).OnCanResize := OnCanResizeStub;
              ExtraData.for_(obj).CanResizeFn := fn;
            end);
end;

procedure TdelQuery.OnCanResizeStub(Sender: TObject; var NewWidth, NewHeight: Integer; var Resize: Boolean);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  event.NewWidth := NewWidth;
  event.NewHeight := NewHeight;
  event.Resize := Resize;
  ExtraData.for_(Sender).CanResizeFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnCanResize <> nil then
      ExtraData.for_(Sender).OldOnCanResize(Sender, NewWidth, NewHeight, Resize);
end;

function TdelQuery.onresize(fn: TdqEventHandler): TdelQuery;

  { fn will be triggered when the control is resized }

begin
  Result := Self;
  Self.each(procedure(obj: TControl)
            begin
              ExtraData.for_(obj).OldOnResize := TWinControlAccessHack(obj).OnResize;
              TWinControlAccessHack(obj).OnResize := OnResizeStub;
              ExtraData.for_(obj).ResizeFn := fn;
            end);
end;

procedure TdelQuery.OnResizeStub(Sender: TObject);
var event: dqevent;
begin
  event.Sender := Sender as TControl;
  ExtraData.for_(Sender).ResizeFn(event);
  if not event.prevent then
    if @ExtraData.for_(Sender).OldOnResize <> nil then
      ExtraData.for_(Sender).OldOnResize(Sender);
end;

function TdelQuery.toggle(fn, fn2: TdqEventHandler): TdelQuery;

  { Toggle among fn and fn2 calls every other click. }

var i: integer;
begin
  Result := self;
  i := 0;
  self.click(procedure(event: dqevent)
             begin
               if i mod 2 = 0 then
                 fn(event)
               else
                 fn2(event);
               inc(i);
             end);
end;

function TdelQuery.Animate(props: array of const;
                           duration: string; easing:
                           TdqEasingFunc = nil;
                           callback: TdqAnimationDoneCallback = nil): TdelQuery;

  { props: The parameters of the control that will be affected by the effect.

      Examples:
      ['height', '+=200'] -> increases the height property by 200px, relative to its current value.
      ['height', '+=200'] -> decreases the height property by 200px, relative to its current value.
      ['height', '200']   -> increases or decreases the height until it reaches 200px exactly.
      ['height', '-=200',
      'width', '+=100']    -> decrease control width by 200 px while incresing width by 100.

    duration: Total time that animation effect occurs. in milliseconds.
              pass 'slow' for 600ms; 'default' for 400ms; 'fast' for 200ms; (inherited from jQuery)
    easing: The easing effect function.
    callback: Called when the animation is completed.

  * Further ref. http://docs.jquery.com/Effects/animate#paramsdurationeasingcallback }

var
  props_: TStringArray;
begin
  Result := Self;
  props_ := StrArrayFromOpenArray(props);
  Self.each( procedure(obj: TControl)
             begin
               DoAnimate(obj, props_, duration, easing, callback);
             end );
end;

{ -------------------------------------------------------------------------------------------------------------------- }
{ The selectors..

  dq(TButton, form1) -> selects all TButton instances in form1.
  dq(TButton, form1, True) -> selects all TButton instances in form1 recursively.
  dq(button1) -> selects the button1 control only.
  dq([button1, button2, button3]) -> selects controls given in the array.

  jQuery equivalent: $(...), jQuery(...) }

function dq(aClass: TClass; Parent: TWinControl; Greedy: boolean): TdelQuery;

  { Selects all instances of control class "aClass" in "Parent".
    Unless "Greedy" is True only direct children are selected. }

  procedure Recurse(aControl: TWinControl);
  var
    i: integer;
    control: TWinControl;
  begin // todo: refactor
    for i := 0 to aControl.ControlCount - 1 do
    begin
      if (aControl.Controls[i] is aClass) then
      begin
        control := (aControl.Controls[i] as aClass) as TWinControl;
        Result.AddControl(control);
      end;
      if (aControl.Controls[i] is TWinControl) and ((aControl.Controls[i] as TWinControl).ControlCount > 0) then
        Recurse(aControl.Controls[i] as TWinControl);
    end;
  end;

var
  control: TControl;
  i: integer;
begin
  Result := TdelQuery.Create;
  if not Greedy then
    for i := 0 to Parent.ControlCount - 1 do
    begin
      if not(Parent.Controls[i] is aClass) then CONTINUE;
      control := (Parent.Controls[i] as aClass) as TControl;
      Result.AddControl(control);
    end
  else
    Recurse(Parent);
end;

function dq(aControl: TControl): TdelQuery;

  { Select given control }

begin
  assert(aControl <> nil);
  Result := dq([aControl]);
end;

function dq(Controls: array of TControl): TdelQuery; overload;

  { Select given controls }

var
  c: TControl;
begin
  Result := TdelQuery.Create;
  for c in Controls do
  begin
    assert(c <> nil);
    Result.AddControl(c);
    ExtraData.for_(c).fDQ := Result;
  end;
end;

initialization
  ExtraDataRegistry := TDictionary<TObject, ExtraData>.Create;
finalization
  FreeAndNil(ExtraDataRegistry);
end.


